"""

"""
from pyglottolog.admin_commands import (
    isoretirements, langindex, updatelinks, updatemetadata, updatesources, updatemacroareas,
)


def run(args):  # pragma: no cover
    isoretirements.run(args)
    langindex.run(args)
    updatelinks.run(args)
    updatemetadata.run(args)
    updatesources.run(args)
    updatemacroareas.run(args)
